#include "ssrf.h"
/* 

driver routines for SSRFPACK

$Id: ssrf_driver.c,v 1.9 2004/03/16 01:52:33 becker Exp becker $

*/

/* 

initialize parameters

*/
void init_sp_structure(struct ssrf_par **sp)
{
  static my_boolean init = FALSE;
  if(init){
    fprintf(stderr,"init_sp_structure: error: should only be called once\n");
    exit(-1);
  }
  *sp = (struct ssrf_par *)calloc(1,sizeof(struct ssrf_par));
  if(!*sp)
    MEMERROR("init_sp_structure");
  (*sp)->istart = 1;
  /* so that realloc works */
  /* float pointers */
  (*sp)->sigma = (*sp)->x = (*sp)->y = (*sp)-> z = NULL;
  /* integer pointers */
  (*sp)->list = (*sp)->lptr = (*sp)->lend = NULL;
  /* default interpolation mode */
  (*sp)->imode = 0;		    
  /* this looks better, normally */
  (*sp)->variable_tension = TRUE;
  /* tolerance for getsig */
  (*sp)->getsig_tol = 1e-4;
  /* start number of iterations for global gradient */
  (*sp)->itgs = 3;
  /* tolerance for gradient */
  (*sp)->getgrad_tol = 1e-5;
  /* max number of iterations for global gradient */
  (*sp)->maxit = 8;
  /* operational flags */
  (*sp)->cart_coord_init = (*sp)->triangulation_init = 
    (*sp)->sigma_init = FALSE;
  init = TRUE;
}

/* 

driver for spherical interpolation routine

INPUT: 

rlon[n], rlat[n]: locations in degrees (unchanged on output) for values: f[n] 

loni and lati, locations at which to interpolate, given in degrees

df: should be set to NULL initially, will be (re-)intialized with local 
    gradients for imode=1
    (for imode == 2, will hold global)
sp: interpolation parameter structure, holds, for instance,

    imode: the mode of interpolation
    0: linear 
    1: c-1 local gradient
    2: c-1 global gradient

reinit: reinitialize interpolation structures

OUTPUT: 

fi, interpolated value at loni and lati, given in degrees
df[n*3]: local gradients, if imode == 0


rlon,rlat, and f might be resorted on output, see reshuffle


if check_for_duplicates is set, will look for doubly specified points
and reduce n accordingly

Thorsten Becker (thwbecker@post.harvard.edu)

$Id: ssrf_driver.c,v 1.9 2004/03/16 01:52:33 becker Exp becker $


*/

/* #define SUPER_DEBUG */
void ssrf_driver(COMP_PRECISION *rlon,COMP_PRECISION *rlat,
		 COMP_PRECISION *f,COMP_PRECISION **df,
		 int *n, COMP_PRECISION loni, COMP_PRECISION lati,
		 COMP_PRECISION *fi,struct ssrf_par *sp,
		 my_boolean reinit,my_boolean randomize,
		 my_boolean check_for_duplicates,
		 my_boolean verbose)
{
  COMP_PRECISION *work,*rlon_rad,*rlat_rad,dsmax,dgmax,diff;
  static my_boolean init = FALSE, warn = FALSE;
  static COMP_PRECISION pif=0.0174532925199433;
  int *iwk,ierr,i,j,igrad_init,ivar_ten,n_reshuffle,n3,
    n6m12,nit,ndup,nold,modified;
  /* 
     start instructions
  */
  if(*n < 3){
    fprintf(stderr,"ssrf_driver: error: on input n: %i, needs to be larger 3\n",
	    *n);
    exit(-1);
  }
  /* size parameters */
  n3 = *n * 3;
  n6m12 = 6 * (*n) - 12;
#ifdef SUPER_DEBUG 
  verbose = TRUE;
  ssrf_print_parameters(sp,*n,stderr);
#endif
  modified=0;
  if((!init)||(reinit)){
    /* 

    INITIALIZATION PHASE
    
    this only gets evaluated once, unless reinit is set to TRUE

    */
    /* 
       allocate space for cartesian coordinates, to be kept
    */
    my_vecrealloc(&sp->x,*n,"ssrf_driver: init: 1");
    my_vecrealloc(&sp->y,*n,"ssrf_driver: init: 1");
    my_vecrealloc(&sp->z,*n,"ssrf_driver: init: 1");
    /* local */
    my_vecalloc(&rlon_rad,*n,"ssrf_driver: init: 1");
    my_vecalloc(&rlat_rad,*n,"ssrf_driver: init: 1");
    /* 
       convert the input locations  
    */
    /* 
       convert lon lat to radians 
    */
    copy_vector(rlon,rlon_rad,*n);scale_vector(rlon_rad,pif,*n);
    copy_vector(rlat,rlat_rad,*n);scale_vector(rlat_rad,pif,*n);
    /* 
       call the SSRF routine to convert all point locations to
       cartesian format
    */
    trans(n,rlat_rad,rlon_rad,sp->x,sp->y,sp->z);	
    sp->cart_coord_init = TRUE;
    /* free the radian coordinate arrays, not needed anymore */
    free(rlon_rad);free(rlat_rad);
    nold = *n;
    ndup = 0;
    if(check_for_duplicates){
      if(verbose)
	fprintf(stderr,"ssrf_driver: running duplicates check\n");
      /* 
	 run a check for duplicate nodes
      */
      for(i=0;i < *n;i++)
	for(j=0;j < i;j++){
	  if(ssrf_nodes_are_close(sp->x[i],sp->y[i],sp->z[i],
				  sp->x[j],sp->y[j],sp->z[j],5e-4)){
	    if((diff=fabs(f[i]-f[j])) > 1e-5){
	      if(warn){
		/* error, different values */
		fprintf(stderr,"ssrf_driver: location %g,%g (%g,%g) specified twice\n",
			rlon[i],rlat[i],rlon[j],rlat[j]);
		fprintf(stderr,"ssrf_driver: cartesian: %g,%g,%g\n",
			sp->x[i],sp->y[i],sp->z[i]);
		fprintf(stderr,"ssrf_driver: cartesian: %g,%g,%g\n",
			sp->x[j],sp->y[j],sp->z[j]);
		fprintf(stderr,"ssrf_driver: warning: different values specified: %g and %g (diff: %g), averaging\n",
			f[i],f[j],diff);
	      }
	      f[i] = (f[i]+f[j])/2.0;
	      f[j] = f[i];
	      modified++;
	    }
	    /* same value specified twice, remove node */
	    ssrf_delete_node(i,&rlon,&rlat,&sp->x,&sp->y,&sp->z,&f,n);
	    i--;		/* step the count one down */
	    ndup++;
	  }
	}
    }
    if((verbose)&&(ndup))
      fprintf(stderr,"ssrf_driver: deleted %i duplicate nodes out of %i, now: %i\n",
	      ndup,nold,*n);
    if((verbose)&&(modified))
      fprintf(stderr,"ssrf_driver: averaged %i nodes\n",modified);
    /* 

       get triangulation, and keep it in storage

    */
    my_ivecrealloc(&sp->list,n6m12,"ssrf_driver: init: 2");
    my_ivecrealloc(&sp->lptr,n6m12,"ssrf_driver: init: 2");
    my_ivecrealloc(&sp->lend,*n,"ssrf_driver: init: 2");
    /* 
       local work arrays 
    */
    my_ivecalloc(&iwk,2*(*n),"ssrf_driver: init: 2");
    my_vecalloc(&work,*n,"ssrf_driver: init: 2");
    if(randomize){
      /* 
	 randomize the sequence of data points, this is recommended
	 for roughly regular distributions
      */
      reshuffle(rlon,rlat,f,sp->x,sp->y,sp->z,*n);
    }
    /* 
       triangulate the input data
    */
    ierr=0;
    trmesh(n,sp->x,sp->y,sp->z,sp->list,sp->lptr,sp->lend,
	   &sp->lnew,iwk,(iwk+(*n)),work,&ierr);
    sp->triangulation_init = TRUE;
    n_reshuffle = 0;
    while((ierr == -2)&&(n_reshuffle < 5)){
      /* 

      first three input data points are collinear, reshuffle all of
      the coordinates randomly

      */
      n_reshuffle++;
      if(verbose)
	fprintf(stderr,"ssrf_driver: reshuffle attempt %i\n",n_reshuffle);
      /* 
	 randomly permutate a bunch of vectors 
      */
      reshuffle(rlon,rlat,f,sp->x,sp->y,sp->z,*n);
      /* 
	 re-try the triangulation 
      */
      ierr=0;
      trmesh(n,sp->x,sp->y,sp->z,sp->list,sp->lptr,sp->lend,
	     &sp->lnew,iwk,(iwk+(*n)),work,&ierr);
    }
    if(ierr != 0){
      fprintf(stderr,"ssrf_driver: init: error code %i in triangulation part\n",
	      ierr);
      if(ierr == -1)
	fprintf(stderr,"ssrf_driver: not enough points, n: %i\n",*n);
      else if(ierr == -2){
	fprintf(stderr,"ssrf_driver: first three points are collinear, even after several reshuffles\n");
	for(i=0;i<3;i++)
	  fprintf(stderr,"ssrf_driver: node %i: x: %11g y: %11g z: %11g\n",
		  i+1,sp->x[i],sp->y[i],sp->z[i]);
      }else{
	fprintf(stderr,"ssrf_driver: duplicates were found\n");
      }
      exit(-1);
    }
    sp->istart = 1;		/* starting node guess */
    free(iwk);free(work);
    /* triangulation done */

    /* 
       
    tension factors
    
    */    
    my_vecrealloc(&sp->sigma,n6m12,"ssrf_driver: init: 3");
    for(i=0;i < n6m12;i++)
      /* initialize tension factors */
      sp->sigma[i] = 0.0;
    sp->sigma_init = TRUE;
    /*
      commented this out, all sigmas are initialized as zero
      if(sp->imode > 0) 
      sp->sigma[0] = 0.0;
      if((sp->imode>0) && (sp->variable_tension)){
      j = sp->lnew-2;
      for(i=0;i < j;i++)
      sp->sigma[i] = 0.0;
      }
    */
    switch(sp->imode){
    case 0:
      /* no further computations needed */
      break;
    case 1:
      /* 
	 C-1 interpolation (INTRC1) with local gradients GRADL. 
      */
      /* compute gradients */
      my_vecrealloc(df,n3,"ssrf_driver: init: 4");
      for(i=0,j=1;i<(*n);i++,j++){
	/* 
	   compute the local, cartesian gradient at node i,
	   sp->df[i*3+0,1,2]. pass node number, j=i+1, in 1..n
	   style
	*/
	ierr=0;
	gradl(n,&j,sp->x,sp->y,sp->z,f,
	      sp->list,sp->lptr,sp->lend,(*df+i*3),&ierr);
	if(ierr < 0){
	  PE("ssrf_driver: init: gradl error for mode 1");
	  exit(-1);
	}
      }
      if(sp->variable_tension){
	/* 
	   Compute tension factors SIGMA (GETSIG)
	*/
	ierr=0;
	getsig(n,sp->x,sp->y,sp->z,f,sp->list,sp->lptr,sp->lend,
	       *df,&sp->getsig_tol,sp->sigma,&dsmax,&ierr);
	if(ierr < 0){
	  PE("ssrf_driver: init: getsig error for mode 1");
	  exit(-1);
	}
      }
      break;
    case 2:
      /* 
	 C-1 interpolation with global gradients GRADG
      */
      my_vecrealloc(df,n3,"ssrf_driver: init: 4");
      /* initialize the gradient  */
      for(i = 0;i < n3;i++)
	*(*df+i) = 0.0;
      if(!sp->variable_tension)
	sp->itgs = 1;
      ivar_ten = 0;		/* for first run */
      /* 
	 0: single tension 
	 1: variable tension
      */
      for(i=0;i < sp->itgs;i++){
	dgmax = sp->getgrad_tol; /* 
				    will be modified by gradg
				 */
	nit = sp->maxit;ierr=0;
	gradg(n,sp->x,sp->y,sp->z,f,
	      sp->list,sp->lptr,sp->lend,&ivar_ten,
	      sp->sigma,&nit,&dgmax,*df,&ierr);
	if(verbose)
	  fprintf(stderr,"ssrf_driver: gradg call %2i: dgmax: %.5e nit: %2i of %2i\n",
		  i+1,dgmax,nit,sp->maxit);
	if(ierr < 0){
	  fprintf(stderr,"ssrf_driver: init: gradg error for mode 2, error code: %i\n",
	     ierr);
	  if(ierr == -3)
	    PE("ssrf_driver: init: gradg: duplicates were found");
	  exit(-1);
	}
	if(sp->variable_tension){
	  ivar_ten = 1;ierr=0;
	  getsig(n,sp->x,sp->y,sp->z,f,
		 sp->list,sp->lptr,sp->lend,
		 *df,&sp->getsig_tol,sp->sigma,&dsmax,&ierr);
	  if(verbose)
	    fprintf(stderr,"ssrf_driver: getsig returned dsmax: %g\n",
		    dsmax);
	}
      }
      break;
    default:
      fprintf(stderr,"ssrf_driver: init: interpolation mode %i undefined\n",
	      sp->imode);
      exit(-1);
      break;
    }
    init = TRUE;
#ifdef SUPER_DEBUG
    ssrf_print_parameters(sp,*n,stderr);
#endif
    if(verbose)
      fprintf(stderr,"ssrf_driver: init: initialization phase done\n");
    /* 
       end init phase 
    */
  }
  /* 

  BEGIN INTERPOLATION PHASE
  
  convert the input point location to radians 
  
  */
  loni *= pif;
  lati *= pif;
  switch(sp->imode){
  case 0:
    /* 
       obtain linear interpolation 
    */
    ierr=0;
    intrc0(n,&lati,&loni,sp->x,sp->y,sp->z,f,
	   sp->list,sp->lptr,sp->lend,&sp->istart,fi,&ierr);
    if(ierr < 0){
      fprintf(stderr,"ssrf_driver: int: error %i in intrc0 routine\n",ierr);
      exit(-1);
    }
    break;     
  case 1:
  case 2:
    /* 
       c-1 smoothed with local or global gradients 
    */
    /* set some flags */
    igrad_init = 1;			/* gradients were precalculated */
    ivar_ten = (sp->variable_tension)?(1):(0);
    /* 
       call the interpolation routine 
    */
    ierr=0;
    intrc1(n,&lati,&loni,sp->x,sp->y,sp->z,f,sp->list,sp->lptr,
	   sp->lend,&ivar_ten,sp->sigma,&igrad_init,*df,
	   &sp->istart,fi,&ierr);
    if(ierr < 0){
      fprintf(stderr,"ssrf_driver: int: error %i in intrc1 routine\n",ierr);
      exit(-1);
    }
    break;
  default:
    fprintf(stderr,"ssrf_driver: int: interpolation mode %i undefined\n",
	    sp->imode);
    exit(-1);
    break;
  }
}
/* 


randomly permutate five x[n] dimensional vectors 


*/
void reshuffle(COMP_PRECISION *x1, COMP_PRECISION *x2,
	       COMP_PRECISION *x3, COMP_PRECISION *x4, 
	       COMP_PRECISION *x5, COMP_PRECISION *x6, int n)
{
  COMP_PRECISION *rand;
  static long seed = -1;
  int *index,i;
  /* get some arrays */
  my_vecalloc(&rand,n,"reshuffle: rand");
  my_ivecalloc(&index,n,"reshuffle: index");
  /* 
     get the random numbers 
  */
  if(seed == -1)
    ran1(&seed);		/* initialize */
  /* obtain vector with n random numbers */
  for(i=0;i<n;i++)
    rand[i] = (COMP_PRECISION) ran1(&seed);
  /* sort according to random numbers to random (pass numrec style, return 
     0..n-1 index) */
  indexx(n,(rand-1),(index-1));
  free(rand);
  /* 
     reassign the vectors
  */
  reassign_vector(x1,index,n);
  reassign_vector(x2,index,n);
  reassign_vector(x3,index,n);
  reassign_vector(x4,index,n);
  reassign_vector(x5,index,n);
  reassign_vector(x6,index,n);
  free(index);
}
/* 

ran1 generator from numerical recipes 


*/
#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-15
#define RNMX (1.0-EPS)

double ran1(long *idum)
{
  int j;
  long k;
  static long iy=0;
  static long iv[NTAB];
  double temp;
  
  if (*idum <= 0 || !iy) {
    if (-(*idum) < 1) *idum=1;
    else *idum = -(*idum);
    for (j=NTAB+7;j>=0;j--) {
      k=(*idum)/IQ;
      *idum=IA*(*idum-k*IQ)-IR*k;
      if (*idum < 0) *idum += IM;
      if (j < NTAB) iv[j] = *idum;
    }
    iy=iv[0];
  }
  k=(*idum)/IQ;
  *idum=IA*(*idum-k*IQ)-IR*k;
  if (*idum < 0) *idum += IM;
  j=iy/NDIV;
  iy=iv[j];
  iv[j] = *idum;
  if ((temp=AM*iy) > RNMX) return RNMX;
  else return temp;
}
#undef IA
#undef IM
#undef AM
#undef IQ
#undef IR
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX

/*
   sort numerical recipes style arrays x[1...n], passed from regular C
   as indexx(...,(x-1),(ix-1)), and gives back indices in 0 ... n-1 fashion

   i.e. pass both arr and indx with (...-1) from normal c style
   programs

*/

#define NUMREC_MINDEXX 7
#define NUMREC_NSTACK 50
#define NUMREC_ISWAP(a,b) {itemp=(a);(a)=(b);(b)=itemp;}


void indexx(int n,COMP_PRECISION *arr,int *indx)
{
  int i,indxt,ir=n,itemp,j,k,l=1;
  int jstack=0,*istack;
  COMP_PRECISION a;
  
  my_ivecalloc(&istack,NUMREC_NSTACK+1,"indexx");

  for (j=1;j<=n;j++) 
    indx[j]=j;
  for (;;) {
    if (ir-l < NUMREC_MINDEXX) {
      for (j=l+1;j<=ir;j++) {
	indxt=indx[j];
	a=arr[indxt];
	for (i=j-1;i>=1;i--) {
	  if (arr[indx[i]] <= a) break;
	  indx[i+1]=indx[i];
	}
	indx[i+1]=indxt;
      }
      if (jstack == 0) break;
      ir=istack[jstack--];
      l=istack[jstack--];
    } else {
      k=(l+ir) >> 1;
      NUMREC_ISWAP(indx[k],indx[l+1]);
      if (arr[indx[l+1]] > arr[indx[ir]]) {
	NUMREC_ISWAP(indx[l+1],indx[ir])
	  }
      if (arr[indx[l]] > arr[indx[ir]]) {
	NUMREC_ISWAP(indx[l],indx[ir])
	  }
      if (arr[indx[l+1]] > arr[indx[l]]) {
	NUMREC_ISWAP(indx[l+1],indx[l])
	  }
      i=l+1;
      j=ir;
      indxt=indx[l];
      a=arr[indxt];
      for (;;) {
	do i++; while (arr[indx[i]] < a);
	do j--; while (arr[indx[j]] > a);
	if (j < i) break;
	NUMREC_ISWAP(indx[i],indx[j])
	  }
      indx[l]=indx[j];
      indx[j]=indxt;
      jstack += 2;
      if (jstack > NUMREC_NSTACK){
	fprintf(stderr,"index: NUMREC_NSTACK too small in indexx.");
	exit(-1);
      }
      if (ir-i+1 >= j-l) {
	istack[jstack]=ir;
	istack[jstack-1]=i;
	ir=j-1;
      } else {
	istack[jstack]=j-1;
	istack[jstack-1]=l;
	l=i;
      }
    }
  }
  free(istack);
  /* go to 0 ... n-1 notation */
  for(i=1;i <= n;i++)
    indx[i]--;
}
#undef NUMREC_MINDEXX
#undef NUMREC_NSTACK
#undef NUMREC_ISWAP

void reassign_vector(COMP_PRECISION *x, int *index, int n)
{
  COMP_PRECISION *xtmp;
  int i;
  my_vecalloc(&xtmp,n,"reassign_vector");
  copy_vector(x,xtmp,n);	/* save x */
  for(i=0;i<n;i++)
    x[i] = xtmp[index[i]];
  free(xtmp);
}
COMP_PRECISION ssrf_norm(COMP_PRECISION *x,int n)
{
  int i;
  double sum = 0.0;
  for(i=0;i<n;i++)
    sum += (double)x[i] * (double)x[i];
  return (COMP_PRECISION)sqrt(sum);
}
unsigned long  ssrf_inorm(int *ix,int n)
{
  int i;
  unsigned long sum = 0;
  for(i=0;i<n;i++)
    sum += (unsigned long) ix[i] * ix[i];
  return (unsigned long) sqrt((COMP_PRECISION)sum);
}
/* 
   
output of all parameters for debugging purposes

 */
void ssrf_print_parameters(struct ssrf_par *sp,int n,
			   FILE *out)
{
  int n6m12;
  n6m12 = n*6-12;
  
  fprintf(out,"ssrf_para: cart_init: %i tri_init: %i sig_init: %i\n",
	  sp->cart_coord_init,sp->triangulation_init,
	  sp->sigma_init);
  fprintf(out,"ssrf_para: imode: %i maxit: %i itgs: %i\n",
	  sp->imode,sp->maxit,sp->itgs);
  fprintf(out,"ssrf_para: tolerance: sig: %11g grad: %11g\n",
	  sp->getsig_tol,sp->getgrad_tol);
  if(sp->cart_coord_init)
    fprintf(out,"ssrf_para: |x|: %11g |y|: %11g |z|: %11g\n",
	    ssrf_norm(sp->x,n),ssrf_norm(sp->y,n),ssrf_norm(sp->z,n));
  if(sp->triangulation_init){
    fprintf(out,"ssrf_para: istart: %i lnew: %i\n",
	    sp->istart,sp->lnew);
    fprintf(out,"ssrf_para: |list|: %li |lptr|: %li |lend|: %li\n",
	    ssrf_inorm(sp->list,n6m12),ssrf_inorm(sp->lptr,n6m12),
	    ssrf_inorm(sp->lend,n));
  }
  if(sp->sigma_init)
    fprintf(out,"ssrf_para: |sigma|: %11g\n",
	    ssrf_norm(sp->sigma,n));
}

my_boolean ssrf_nodes_are_close(COMP_PRECISION x1,COMP_PRECISION y1,COMP_PRECISION z1,
				COMP_PRECISION x2,COMP_PRECISION y2,COMP_PRECISION z2,
				COMP_PRECISION eps)
{
  if((fabs(x1-x2)<eps)&&(fabs(y1-y2)<eps)&&(fabs(z1-z2)<eps)){
    return TRUE;
  }  else {
    return FALSE;
  }

}
void ssrf_delete_node(int i,
		      COMP_PRECISION **x1,COMP_PRECISION **x2,
		      COMP_PRECISION **x3,COMP_PRECISION **x4,
		      COMP_PRECISION **x5,COMP_PRECISION **x6,
		      int *n)
{
  int n1;
  n1 = *n - 1;
  if(i != n1){
    /* swap the last items of each vector into the location which is to be deleted */
    *(*x1+i) = *(*x1+n1);
    *(*x2+i) = *(*x2+n1);
    *(*x3+i) = *(*x3+n1);
    *(*x4+i) = *(*x4+n1);
    *(*x5+i) = *(*x5+n1);
    *(*x6+i) = *(*x6+n1);
  }
  /* reduce the vector length */
  *n = *n - 1;
  *x1=(COMP_PRECISION *)realloc(*x1,sizeof(COMP_PRECISION)*(*n));
  *x2=(COMP_PRECISION *)realloc(*x2,sizeof(COMP_PRECISION)*(*n));
  *x3=(COMP_PRECISION *)realloc(*x3,sizeof(COMP_PRECISION)*(*n));
  *x4=(COMP_PRECISION *)realloc(*x4,sizeof(COMP_PRECISION)*(*n));
  *x5=(COMP_PRECISION *)realloc(*x5,sizeof(COMP_PRECISION)*(*n));
  *x6=(COMP_PRECISION *)realloc(*x6,sizeof(COMP_PRECISION)*(*n));
}

COMP_PRECISION ssrf_sqr(COMP_PRECISION x)
{return (x*x);}
